def to_utf8_bytes(text):
    try:
        return text.encode("utf-8", "strict")
    except UnicodeError as err:
        raise TypeError(
            "'{}' is cannot be converted to UTF-8".format(text)
        ) from err



from __future__ import annotations

import math
import re
from decimal import Decimal, ROUND_HALF_UP
from dateutil.parser._parser import ParserError
from typing import Dict, Hashable, Union
import json
import numpy
import pandas
from pandas import Series

from .utils import to_utf8_bytes
from .errors import InvalidRedshiftType

Dtype = Union[str, "RedshiftType"]
DtypeArg = Union[Dtype, Dict[Hashable, Dtype]]

_TYPE_REGEX = re.compile(r"^([a-zA-Z0-9 ]*)(\(([0-9, ]*?)\))?$")


def get_redshift_type(type_str):
    m = _TYPE_REGEX.match(type_str)
    if not m:
        raise InvalidRedshiftType(
            "Redshift type not found for '{}'".format(type_str)
        )

    type_name = m.group(1)
    type_args = m.group(3)

    type_name = type_name.upper().strip()

    type_dict = {
        "SMALLINT": SmallInt,
        "INT2": SmallInt,
        "INTEGER": Integer,
        "INT": Integer,
        "INT4": Integer,
        "BIGINT": BigInt,
        "INT8": BigInt,
        "DECIMAL": Numeric,
        "NUMERIC": Numeric,
        "REAL": Real,
        "FLOAT4": Real,
        "DOUBLE PRECISION": DoublePrecision,
        "FLOAT8": DoublePrecision,
        "FLOAT": DoublePrecision,
        "BOOLEAN": Boolean,
        "BOOL": Boolean,
        "CHAR": Char,
        "CHARACTER": Char,
        "NCHAR": Char,
        "BPCHAR": BPChar,
        "VARCHAR": VarChar,
        "CHARACTER VARYING": VarChar,
        "NVARCHAR": VarChar,
        "TEXT": Text,
        "DATE": Date,
        "TIMESTAMP": TimeStamp,
        "TIMESTAMP WITHOUT TIME ZONE": TimeStamp,
        "TIMESTAMPTZ": TimeStampTz,
        "TIMESTAMP WITH TIME ZONE": TimeStampTz,
        "TIME": Time,
        "TIME WITHOUT TIME ZONE": Time,
        "TIMETZ": TimeTz,
        "TIME WITH TIME ZONE": TimeTz,
        "GEOMETRY": Geometry,
        "SUPER": Super,
    }

    if type_name not in type_dict:
        raise InvalidRedshiftType(
            "Redshift type not found for '{}'".format(type_str)
        )
    redshift_type = type_dict[type_name]
    if type_args:
        type_args = [int(elm.strip()) for elm in type_args.split(",")]
    else:
        type_args = []

    return redshift_type(*type_args)


class RedshiftType(object):
    """An abstracttype for Redshift types.

    Each type has encoder and decoder.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        type, for ease of dealing.

    """

    _ESCAPES = [
        ("\\", "\\\\"),
        ("'", "\\'"),
        ("\n", "\\n"),
        ("\t", "\\t"),
        ("\b", "\\b"),
        ("\f", "\\f"),
    ]

    def _check(self, text, ubytes):
        pass

    def _encode_text(self, text):
        if pandas.isnull(text) or pandas.isna(text):
            return "NULL"

        ubytes = to_utf8_bytes(str(text))

        encoded_text = ubytes.decode("utf-8")

        self._check(encoded_text, ubytes)

        encoded_text = "\n".join(encoded_text.splitlines())

        for old, new in self._ESCAPES:
            encoded_text = encoded_text.replace(old, new)

        return "'{}'".format(encoded_text)

    def encode(self, col: Series) -> Series:
        """Encode objects stored in a column for pandas.DataFrame to
        ``str``-typed Redshift notations, which are used in DMLs.

        First, values are casted to string. Next, character encoding is
        changed to ``utf-8``, which Redshift supports as a multibyte
        character set. Next, strings are checked in terms of length or
        multibyte characters to avoid errors when running ``INSERT``
        statements. Then, escapes are replaced. Finally, the string is quoted.

        Parameters
        ----------
        col : pandas.Series
            The column storing original objects in pandas.DataFrame.

        Returns
        -------
        encoded_col : pandas.Series
            Column storing Redshift notations.

        """

        encoded_col = col.fillna(numpy.nan)
        encoded_col = encoded_col.map(self._encode_text)

        return encoded_col

    def decode(self, col: Series) -> Series:
        """Decode response from Redshift data api to Python ojects. See
        comments on each Redshift type class to confirm what type or class
        is used.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing Python objects.

        """

        return col

    def __str__(self):
        return self.__redshift_name__


class DoublePrecision(RedshiftType):
    """A type for Redshift ``DOUBLE PRECISION`` type.

    This type is decoded to numpy ``float64`` type.

    The encoder for this type accepts any types which are able to
    casted to numpy ``float64`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy float type, for ease of dealing.

    """

    __np_type__ = "float64"
    __redshift_name__ = "DOUBLE PRECISION"
    __min_abs__ = 2.22507385850721e-308
    __max_abs__ = 1.79769313486231e308
    __to_be_checked__ = True

    def _check_range(self, val):
        if pandas.isna(val) or val == 0.0:
            return val
        val_abs = abs(val)
        if val_abs < self.__min_abs__ or self.__max_abs__ < val_abs:
            raise TypeError(
                "'{}' is out of range for type '{}'".format(val, str(self))
            )
        return val

    def encode(self, col: Series) -> Series:
        """Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.

        First, values are casted to numpy float type. Next, value
        range are checked to avoid overflow or underflow errors
        when running ``INSERT`` statements. Finally, the numeric
        types are casted to str.

        Parameters
        ----------
        col : pandas.Series
            The column storing original objects in pandas.DataFrame.

        Returns
        -------
        encoded_col : pandas.Series
            Column storing Redshift notations.

        """

        encoded_col = col.astype(self.__np_type__)

        encoded_col = encoded_col.fillna(numpy.nan)
        if self.__to_be_checked__:
            encoded_col.map(self._check_range)
        encoded_col = encoded_col.replace([numpy.nan], ["NULL"])
        encoded_col = encoded_col.astype(str)

        return encoded_col

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str or float types. This decoder will map these raw values to
        the proper numpy float type, for ease of dealing.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing numpy float values.

        """

        return col.astype(self.__np_type__)


class Real(DoublePrecision):
    """A type for Redshift ``REAL`` type.

    This type is decoded to numpy ``float64`` type since deciaml
    inaccuracy is observed in case of using numpy ``float32``.

    The encoder for this type accepts any values which are able to
    casted to numpy ``float64`` type and do not cause overflow or
    underflow for Redshift ``REAL`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy float type, for ease of dealing.

    """

    __redshift_name__ = "REAL"
    __min_abs__ = 1.1755e-38
    __max_abs__ = 3.40282e38


class Numeric(DoublePrecision):
    """A type for Redshift ``DECIMAL`` type.

    In this library, the alias ``NUMERIC`` is used instead to avoid
    conflict with Python ``decimal.Decimal`` type.

    There are not any fixed point types in numpy. This made us
    develop the decoder to cast values from Redshift Data API to
    Python ``decimal.Decimal``. Hense, the output for the decoder
    looks ``object``-type Series.

    The encoder for this type accepts any values which are able to
    casted to numpy ``float128`` type and do not cause overflow
    for the decimal with the specific precision and scale.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        ``decimal.Decimal`` type, for ease of dealing.

    """

    __np_type__ = "float128"
    __redshift_name__ = "NUMERIC"

    def __init__(self, precision: int = 18, scale: int = 0):
        """Construct the Redshift ``NUMERIC`` type.

        Parameters
        ----------
        precision :
            the numeric precision for use in DDL ``CREATE TABLE``.

        :param scale: the numeric scale for use in DDL ``CREATE TABLE``.

        """

        if precision != 18 or scale != 0:
            self.__redshift_name__ = "NUMERIC({},{})".format(precision, scale)
        self.__max_abs__ = Decimal(str(math.pow(10.0, precision - scale)))

        self.__exp_to_quantize__ = Decimal(
            "1.{}".format("".join(["0" for i in range(scale)]))
        )

    def _encode_numeric(self, val):
        if pandas.isna(val):
            return "NULL"

        decimal_val = Decimal(str(val)).quantize(
            self.__exp_to_quantize__, rounding=ROUND_HALF_UP
        )

        decimal_val_abs = abs(decimal_val)
        if self.__max_abs__ <= decimal_val_abs:
            raise TypeError(
                "'{}' is out of range for type '{}'".format(
                    decimal_val, str(self)
                )
            )
        return str(decimal_val)

    def encode(self, col: Series) -> Series:
        """Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.

        First, values are casted to numpy.float128 type to avoid
        numeric inaccuracy. Next, numpy.float128 values are converted to
        decimal.Decimal values which are accurate under the conditions of
        ``precision`` and ``scale.`` Then, after checking min/max to avoid
        overflow error, values are casted to str.

        Parameters
        ----------
        col : pandas.Series
            The column storing original objects in pandas.DataFrame.

        Returns
        -------
        encoded_col : pandas.Series
            Column storing Redshift notations.
        """

        encoded_col = col.astype(self.__np_type__)

        encoded_col = encoded_col.fillna(numpy.nan)
        encoded_col = encoded_col.map(self._encode_numeric)
        return encoded_col

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will cast these raw values represented
        in string to ``decimal.Decimal`` objects. To store
        ``decimal.Decimal`` objects, the ``dtype`` for the returned
        pandas.Series looks ``object``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing ``decimal.Decimal`` values.

        """

        def _to_decimal(val):
            if pandas.isna(val):
                return numpy.nan

            return Decimal(val)

        return col.map(_to_decimal)


class Integer(DoublePrecision):
    """A type for Redshift ``INTEGER`` type.

    This type values are decoded to numpy ``int32`` in case NULL is
    not included, and otherwise decoded to pandas ``Int64``, which
    is the nullable integer type.

    The encoder for this type accepts any values which are able to
    casted to numpy ``Int64`` type and do not cause overflow
    for Redshift ``INTEGER`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy int type, for ease of dealing.

    """

    __np_type__ = "Int64"
    __np_type_not_null__ = "int32"
    __redshift_name__ = "INTEGER"
    __min__ = -2147483648
    __max__ = 2147483647

    def _check_range(self, val):
        if pandas.isna(val):
            return numpy.nan
        if val < self.__min__ or self.__max__ < val:
            raise TypeError(
                "'{}' is out of range for type '{}'".format(val, str(self))
            )
        return val

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str or int type. This decoder will cast these raw values to
        numpy int objects.

        If NULL is included, this decoder will use the nullable integer
        type ``Int64``. Otherwise, numpy integer types, which are ``int16``,
        ``int32``, or ``int64``, are used.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing numpy int values.

        """

        if len(col) != col.count():
            return col.astype(self.__np_type__)
        return col.astype(self.__np_type_not_null__)


class SmallInt(Integer):
    """A type for Redshift ``SMALLINT`` type.

    This type values are decoded to numpy ``int16`` in case NULL is
    not included, and otherwise decoded to pandas ``Int64``, which
    is the nullable integer type.

    The encoder for this type accepts any values which are able to
    casted to numpy ``Int64`` type and do not cause overflow
    for Redshift ``SMALLINT`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy int type, for ease of dealing.

    """

    __np_type_not_null__ = "int16"
    __redshift_name__ = "SMALLINT"
    __min__ = -32768
    __max__ = 32767


class BigInt(Integer):
    """A type for Redshift ``BIGINT`` type.

    This type values are decoded to numpy ``int64`` in case NULL is
    not included, and otherwise decoded to pandas ``Int64``, which
    is the nullable integer type.

    The encoder for this type accepts any values which are able to
    casted to numpy ``Int64`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy int type, for ease of dealing.

    """

    __np_type_not_null__ = "int64"
    __redshift_name__ = "BIGINT"
    __to_be_checked__ = False


class Boolean(Integer):
    """A type for Redshift ``BOOLEAN`` type.

    This type values are decoded to numpy ``bool`` in case NULL is
    not included, and otherwise decoded to pandas ``boolean``, which
    is the nullable integer type.

    The encoder for this type accepts any values which are able to
    casted to numpy ``boolean`` type.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        numpy bool type, for ease of dealing.

    """

    __np_type__ = "boolean"
    __np_type_not_null__ = "bool"
    __redshift_name__ = "BOOLEAN"
    __to_be_checked__ = False

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will cast these raw values to numpy
        boolean objects.

        If NULL is included, this decoder will use the nullable boolean
        type ``boolean``. Otherwise, numpy boolean type ``bool`` is used.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing numpy bool values.

        """

        decoded_col = col.map(
            lambda x: numpy.nan if pandas.isna(x) else (x == "true")
        )
        return super().decode(decoded_col)


class Char(RedshiftType):
    """A type for Redshift ``CHAR`` type.

    This type values are decoded to Python ``str``.

    The encoder for this type accepts strings, but it rejects multibyte
    characters and too long strings.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the Python
        ``str`` type, for ease of dealing.

    """

    __multibyte_is_allowed__ = False
    __redshift_name__ = "CHAR"
    __default_length__ = 1
    __max_length__ = 4096

    def __init__(self, length: int = 0):
        """Construct the Redshift ``CHAR``type.

        Parameters
        ----------
        length :
            Length limitation.

        """

        self.length = self.__default_length__ if length == 0 else length
        if self.length != self.__default_length__:
            if self.length > self.__max_length__:
                raise InvalidRedshiftType(
                    "The length '{}' is too long for '{}'".format(
                        self.length, self.__redshift_name__
                    )
                )
            self.__redshift_name__ = "{}({})".format(
                self.__redshift_name__, length
            )

    def _check(self, text, ubytes):
        if (not self.__multibyte_is_allowed__) and len(text) != len(ubytes):
            raise TypeError("multibyte characters must not be included")
        if len(ubytes) > self.length:
            raise TypeError(
                "'{}' exceeds length ({})".format(text, self.length)
            )


class BPChar(Char):
    """A type for Redshift ``BPCHAR`` type. This type is alias for
    ``CHAR`` type, but the specification about length is different:
    the length is fixed as 256.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the Python
        ``str`` type, for ease of dealing.

    """

    __redshift_name__ = "BPCHAR"
    __default_length__ = 256

    def __init__(self):
        """Construct the Redshift ``BPCHAR`` type."""

        self.length = self.__default_length__


class VarChar(Char):
    """A type for Redshift ``VARCHAR`` type.

    This type values are decoded to Python ``str``.

    The encoder for this type accepts strings. Unlike ``CHAR`` type,
    this type accepts multibyte characters.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the Python
        ``str`` type, for ease of dealing.

    """

    __multibyte_is_allowed__ = True
    __redshift_name__ = "VARCHAR"
    __default_length__ = 256
    __max_length__ = 65535


class Text(VarChar):
    """A type for Redshift ``TEXT`` type. This type is alias for
    ``VARCHAR`` type, but the specification about length is different:
    the length is fixed as 256.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the Python
        ``str`` type, for ease of dealing.

    """

    __redshift_name__ = "TEXT"

    def __init__(self):
        """Construct the Redshift ``TEXT`` type."""

        self.length = self.__default_length__


class TimeStamp(RedshiftType):
    """A type for Redshift ``TIMESTAMP`` type.

    This type values are decoded with ``pandas.to_datetime``.

    The encoder for this type accepts any values which can be
    converted to datetime objects with ``pandas.to_datetime``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        datetime.datetime type, for ease of dealing.

    """

    __redshift_name__ = "TIMESTAMP"
    __dt_format__ = "%Y-%m-%d %H:%M:%S"
    __utc__ = False

    def encode(self, col: Series) -> Series:
        """Encode objects stored in a column for pandas.DataFrame to
        ``str``-typed notations for Redshift DMLs.

        First, values are converted to datetime objects with
        ``pandas.to_datetime``.
        Next, values are converted to text with ``strftime`` and format.
        In the format, quote is included.

        Parameters
        ----------
        col : pandas.Series
            The column storing original objects in pandas.DataFrame.

        Returns
        -------
        encoded_col : pandas.Series
            Column storing Redshift notations.

        """

        def _strftime(obj):
            if pandas.isnull(obj) or pandas.isna(obj):
                return numpy.nan
            if hasattr(obj, "strftime"):
                return obj.strftime(self.__dt_format__)
            return obj

        # to deal with such types that is unable to convert to datetime
        # with ``pandas.to_datetime`` like Timea
        encoded_col = col.map(_strftime)
        try:
            encoded_col = pandas.to_datetime(encoded_col, utc=self.__utc__)
        except ParserError as err:
            raise TypeError("cannot parse to datetime {}".format(str(err)))

        output_format = "'{}'".format(self.__dt_format__)
        encoded_col = encoded_col.dt.strftime(output_format)
        encoded_col = encoded_col.fillna("NULL")

        return encoded_col

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will convert raw values to datetime
        objects with ``pandas.to_datetime``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing datetime.datetime values.

        """

        return pandas.to_datetime(col, errors="coerce", utc=self.__utc__)


class TimeStampTz(TimeStamp):
    """A type for Redshift ``TIMESTAMPTZ`` type.

    This type values are decoded with ``pandas.to_datetime`` and
    option ``utc``.

    The encoder for this type accepts any values which can be
    converted to datetime objects with ``pandas.to_datetime``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        datetime.datetime type, for ease of dealing.

    """

    __redshift_name__ = "TIMESTAMPTZ"
    __dt_format__ = "%Y-%m-%d %H:%M:%S%z"
    __utc__ = True


class Date(TimeStamp):
    """A type for Redshift ``DATE`` type.

    This type values are decoded by converting to datetime objects
    with ``pandas.to_datetime`` and in addition converting to date
    objects by ``datetime.date()``.

    The encoder for this type accepts any values which can be
    converted to datetime objects with ``pandas.to_datetime``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        datetime.date type, for ease of dealing.

    """

    __redshift_name__ = "DATE"
    __dt_format__ = "%Y-%m-%d"

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. First, this decoder converts raw values to datetime
        objects with ``pandas.to_datetime``. Next, datetime objects are
        converted to date objects with ``datetime.date()``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing datetime.date values.

        """

        col = super().decode(col)
        return col.map(lambda dt: dt.date() if not pandas.isna(dt) else dt)


class Time(TimeStamp):
    """A type for Redshift ``TIME`` type.

    This type values are decoded by converting to datetime objects
    with ``pandas.to_datetime`` and in addition converting to time
    objects by ``datetime.time()``.

    The encoder for this type accepts any values which can be
    converted to datetime objects with ``pandas.to_datetime``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        datetime.time type, for ease of dealing.

    """

    __redshift_name__ = "TIME"
    __dt_format__ = "%H:%M:%S"

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will convert raw values to datetime
        objects with ``pandas.to_datetime``. Next, datetime objects are
        converted to time objects with ``datetime.time()``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing datetime.time values.

        """

        col = super().decode(col)
        return col.map(lambda dt: dt.time() if not pandas.isna(dt) else dt)


class TimeTz(TimeStamp):
    """A type for Redshift ``TIMETZ`` type.

    This type values are decoded by converting to datetime objects
    with ``pandas.to_datetime`` and in addition converting to time
    objects by ``datetime.timetz()``.

    The encoder for this type accepts any values which can be
    converted to datetime objects with ``pandas.to_datetime``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        datetime.time type, for ease of dealing.

    """

    __redshift_name__ = "TIMETZ"
    __dt_format__ = "%H:%M:%S%z"
    __utc__ = True

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will convert raw values to datetime
        objects with ``pandas.to_datetime``. Next, datetime objects are
        converted to timetz objects with ``datetime.timetz()``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing datetime.time values.

        """
        col = super().decode(col)
        return col.map(lambda dt: dt.timetz() if not pandas.isna(dt) else dt)


class Geometry(RedshiftType):
    """ "A type for Redshift ``GEOMETRY`` type.

    The decoder or encoder for this type has not been implemented.

    GEOMETRY values can be represented in ``str`` for both query
    response and DMLs.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to Python
        ``str`` type, for ease of dealing.

    """

    __redshift_name__ = "GEOMETRY"


class Super(RedshiftType):
    """A type for Redshift ``SUPER`` type. This type is equivalent
    to Python generic types: ``int``, ``float``, ``str``, ``list``,
    ``dict``, and ``None``.

    This type values are decoded by converting to Python objects
    mainly with ``json.loads``.

    The encoder for this type accepts the six Python types mentioned
    above. The semi-structured data, ``list`` and ``dict`` types,
    are parsed with ``JSON_PARSE`` before ingestion to Redshift.

    This type is sub-class for ``VarChar`` to leverage the private
    method ``_encode_text``.

    Methods
    -------
    encode :
        Encode objects stored in a column for pandas.DataFrame
        to ``str``-typed notations for Redshift DMLs.
    decode :
        Map raw values from Redshift Data API to the proper
        Python object type, for ease of dealing.

    """

    __redshift_name__ = "SUPER"

    def __init__(self):
        """Construct the Redshift ``SUPER`` type."""
        self.length = 0

    def encode(self, col: Series) -> Series:
        """Encode objects stored in a column for pandas.DataFrame to
        ``str``-typed notations for Redshift DMLs.

        The behavior depends on the type for the value. 1. The ``int``
        or ``float`` values are casted to ``str``, 2. ``str`` values
        are transformed with the same logic as ``CHAR``/``VARCHAR``-type
        encoder, 3. ``list`` or ``dict`` values are once transformed as
        ``str`` values are done and the transformed values are set as
        the argument for ``JSON_PARSE``.

        Parameters
        ----------
        col : pandas.Series
            The column storing original objects in pandas.DataFrame.

        Returns
        -------
        encoded_col : pandas.Series
            Column storing Redshift notations.

        """

        def _encode_super(obj):
            if obj is None:
                return "NULL"
            elif type(obj) is int or type(obj) is float:
                if pandas.isna(obj):
                    return "NULL"
                return str(obj)
            elif type(obj) is str:
                return self._encode_text(obj)
            elif type(obj) is dict or type(obj) is list:
                json_str = json.dumps(obj)
                encoded_json = self._encode_text(json_str)
                return "JSON_PARSE({})".format(encoded_json)
            else:
                raise TypeError(
                    "unsupported datatype {} for SUPER type".format(type(obj))
                )

        encoded_col = col.astype("object")
        encoded_col = col.map(_encode_super)
        return encoded_col

    def decode(self, col: Series) -> Series:
        """Raw values in response from Redshift data api are represented
        in str type. This decoder will convert raw values to Python
        objects with ``json.loads``.

        Parameters
        ----------
        col :
            Column storing raw values in response from Redshift Data API.

        Returns
        -------
        col :
            Column storing Python object values.

        """

        def _decode_super(obj):
            if pandas.isna(obj):
                return numpy.nan

            return json.loads(obj)

        return col.map(_decode_super)



from __future__ import annotations

import logging
from typing import cast
from time import sleep
import boto3

from collections import OrderedDict

import numpy
import pandas
from pandas import DataFrame
from pandas.core.dtypes.common import is_dict_like

from .utils import to_utf8_bytes
from .types import get_redshift_type, RedshiftType, DtypeArg
from .errors import (
    InvalidAuthentication,
    QueryFailedError,
    QueryAbortedError,
    TableCreationError,
    MetadataEncodeError,
    DataEncodeError,
    NoColumnError,
)

logger = logging.getLogger(__name__)


def to_redshift(
    dataframe,
    cluster_identifier: str,
    database: str,
    table: str,
    dtype: DtypeArg,
    schema: str = "public",
    db_user: str | None = None,
    secret_arn: str | None = None,
    if_exists: str = "fail",
) -> None:
    """Write a DataFrame to a Redshift table.

    The main method a user calls to export pandas DataFrame contents to
    Amazon Redshift table.

    This method uses the RedshiftDataAPIService for Boto 3 to make requests to
    Amazon Redshift, documented `here
    <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift-data.html#RedshiftDataAPIService.Client.execute_statement>`__.

    The authentication method must be adopted from temporary
    credentials (db_user/database) or AWS Secrets Manager (secret_arn).

    Parameters
    ----------
    dataframe : pandas.DataFrame
        DataFrame to be written to a Amazon Redshift table.
    cluster_identifier : str
        The cluster identifier.
    database : str
        The name of database. This parameter is required when authenticating
        using temporary credentials.
    table : str
        Name of table to be written
    dtype: dict or scalar
        Specifying the datatype for columns. If a dictionary is used, the
        keys should be the column names and the values should be the
        Redshift types or equivalent strings. If a scalar is provided,
        it will be applied to all columns.
    schema, default "public"
        Name of schema where the table is written
    db_user : str, optional
        The database user name. This parameter is required when authenticating
        using temporary credentials.
    secret_arn : str, optional
        The name or ARN of the secret that enables access to the database.
        This parameter is required when authenticating using AWS Secrets
        Manager.
    if_exists : str, default 'skip'
        Behavior when the desination table exists. Value can be one of:

        ``'fail'``
            If table exists, abort the process with error.
    """

    connector = RedshiftConnector(
        cluster_identifier, database, db_user=db_user, secret_arn=secret_arn
    )

    if connector.exists_table(table, schema):
        if if_exists == "fail":
            raise TableCreationError(
                "Could not create the table {} in the schema {} "
                "because it already exists.".format(table, schema)
            )

    dtype = connector.prep_dtype(dataframe, dtype)

    data, columns = connector.encode_data(dataframe, dtype, table, schema)

    connector.create_table(columns, dtype, table, schema)

    connector.load_dataframe(data, columns, dtype, table, schema)


def read_redshift(
    select_sql,
    cluster_identifier: str,
    database: str,
    db_user: str | None = None,
    secret_arn: str | None = None,
) -> DataFrame:
    """Load data from Amazon Redshift using Amazon Redshift Data API

    The main method a user calls to execute a SQL in Amazon Redshift
    and read result into a pandas DataFrame.

    This method uses the RedshiftDataAPIService for Boto3 to make requests to
    Amazon Redshift, documented `here
    <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/redshift-data.html#RedshiftDataAPIService.Client.execute_statement>`__.

    The authentication method must be adopted from temporary credentials
    (db_user/database) or AWS Secrets Manager (secret_arn).

    Parameters
    ----------
    sql : str
        SQL to return data values.
    cluster_identifier : str
        The cluster identifier.
    database : str
        The name of database. This parameter is required when authenticating
        using temporary credentials.
    db_user : str, optional
        The database user name. This parameter is required when authenticating
        using temporary credentials.
    secret_arn : str, optional
        The name or ARN of the secret that enables access to the database.
        This parameter is required when authenticating using AWS Secrets
        Manager.

    Returns
    -------
    result_df : pandas.DataFrame
        DataFrame storing data from Amazon Redshift.
    """

    connector = RedshiftConnector(
        cluster_identifier, database, db_user=db_user, secret_arn=secret_arn
    )

    result_df = connector.to_dataframe(select_sql)

    return result_df


class RedshiftConnector(object):
    client = None

    def __new__(cls, *args, **kwargs):
        if not cls.client:
            cls.client = boto3.client("redshift-data")
        return super(RedshiftConnector, cls).__new__(cls)

    def __init__(
        self, cluster_identifier, database, db_user=None, secret_arn=None
    ):
        self.cluster_identifier = cluster_identifier
        self.db_user = db_user
        self.database = database
        self.secret_arn = secret_arn
        if self.secret_arn is None and self.db_user is None:
            raise InvalidAuthentication(
                "Authentication requires db_user or secret_arn."
            )

    # the specific string to identify no token (no more results)
    _NO_NEXT_TOKEN = ""

    def _has_next_token(self, params):
        # The case "NextToken" is not included means the first iteration
        if "NextToken" not in params:
            return True

        if params["NextToken"] == self._NO_NEXT_TOKEN:
            return False

        return True

    def run_query(self, sql):
        execute_statement_params = {
            "ClusterIdentifier": self.cluster_identifier,
            "Database": self.database,
            "Sql": sql,
        }
        if self.secret_arn is not None:
            execute_statement_params["SecretArn"] = self.secret_arn
        else:
            execute_statement_params["DbUser"] = self.db_user

        result = self.client.execute_statement(**execute_statement_params)
        statement_id = result["Id"]

        logger.debug(
            "Statement ID: {} (sql: '{}') started".format(statement_id, sql)
        )

        return result

    def wait(self, sqls, statement_ids):
        for index in range(len(statement_ids)):
            sql = sqls[index]
            statement_id = statement_ids[index]

            status, description = "", {}
            while status not in {"FINISHED", "FAILED", "ABORTED"}:
                description = self.client.describe_statement(Id=statement_id)
                status = description["Status"]
                sleep(1.0)

            if status == "FAILED":
                raise QueryFailedError(
                    "The following query was failed "
                    "[ID: {} (sql: '{}')]\n({})".format(
                        statement_id, sql, description["Error"]
                    )
                )
            elif status == "ABORTED":
                raise QueryAbortedError(
                    "The following query was stopped by user "
                    "[ID: {} (sql: '{}')]".format(statement_id, sql)
                )

            logger.debug(
                "Statement ID: {} (sql: '{}') finished".format(
                    statement_id, sql
                )
            )

    @staticmethod
    def _decode_data(df, redshift_types):
        """Make the DataFrame's column types align with the Redshift table
        column types.
        Need to work around limited NA value support. Floats are always
        fine, ints must always be floats if there are Null values.
        Booleans are hard because converting bool column with None replaces
        all Nones with false. Therefore only convert bool if there are no
        NA values.
        Datetimes should already be converted to np.datetime64 if supported,
        but here we also force conversion if required.
        """

        for col_name, type_name in redshift_types.items():
            col = df[col_name]
            redshift_type = get_redshift_type(type_name)

            df[col_name] = redshift_type.decode(col)

    def to_dataframe(self, select_sql):
        run_result = self.run_query(select_sql)

        statement_id = run_result["Id"]

        self.wait([select_sql], [statement_id])

        # Now status is "FINISHED"
        get_statement_result_params = {"Id": statement_id}
        redshift_types = OrderedDict()
        data = []
        while self._has_next_token(get_statement_result_params):
            statement_result = self.client.get_statement_result(
                **get_statement_result_params
            )

            if "NextToken" not in get_statement_result_params:
                for col in statement_result.get("ColumnMetadata", []):
                    col_name = col["name"]
                    col_type = col["typeName"]
                    redshift_types[col_name] = col_type

            chunk = [
                tuple(
                    [
                        numpy.nan
                        if "isNull" in cell
                        else list(cell.values())[0]
                        for cell in record
                    ]
                )
                for record in statement_result.get("Records", [])
            ]
            data.extend(chunk)

            get_statement_result_params["NextToken"] = statement_result.get(
                "NextToken", self._NO_NEXT_TOKEN
            )

        # to set approprite columns and dtypes for empty table
        # from_records and astype are necessary
        data = pandas.DataFrame.from_records(
            data, columns=list(redshift_types.keys())
        )

        self._decode_data(data, redshift_types)

        return data

    def prep_dtype(self, dataframe, dtype):
        if not is_dict_like(dtype):
            # error: Value expression in dictionary comprehension has
            # incompatible type "Union[str, RedshiftType,
            # Dict[Hashable, Union[str, RedshiftType]]]";
            # expected type "Union[str, RedshiftType]"
            dtype = {column: dtype for column in dataframe}
        else:
            dtype = cast(dict, dtype)

        for column in dataframe:
            my_type = dtype[column]
            if not isinstance(my_type, RedshiftType):
                my_type = get_redshift_type(my_type)
                dtype[column] = my_type

        return dtype

    def create_table(self, columns, dtype, table, schema):
        encode = self._encode_metadata

        col_def = [
            "{} {}".format(encode(column), dtype[column]) for column in columns
        ]
        create_table_sql = "CREATE TABLE {}.{} ({});".format(
            encode(schema), encode(table), ",".join(col_def)
        )
        run_result = self.run_query(create_table_sql)

        statement_id = run_result["Id"]

        self.wait([create_table_sql], [statement_id])

    _MAX_STATEMENT_SIZE = 100000

    def encode_data(self, dataframe, dtype, table, schema):
        encoded_df = dataframe.copy()
        columns = list(encoded_df.columns)

        if len(columns) == 0:
            raise NoColumnError("Table must have at least one column")

        for column in columns:
            my_type = dtype[column]
            try:
                encoded_df[column] = my_type.encode(encoded_df[column])
            except TypeError as err:
                raise DataEncodeError(
                    "Column '{}': {}".format(column, str(err))
                )

        # to efficiently access to rows in dataframe
        data = encoded_df.to_numpy()

        self._check_row_length(data, columns, table, schema)

        return data, columns

    def _encode_metadata(self, name):
        try:
            uname = to_utf8_bytes(str(name)).decode("utf-8")
        except TypeError as err:
            raise MetadataEncodeError("Identifier {}".format(str(err)))

        if not len(uname):
            raise MetadataEncodeError("Empty table or column name specified")

        nul_index = uname.find("\x00")
        if nul_index >= 0:
            raise MetadataEncodeError("Identifier cannot contain NULs")

        return '"{}"'.format(uname.replace('"', '""'))

    def _check_row_length(self, data, columns, table, schema):
        insert_sql_prefix = self._generate_insert_sql_prefix(
            columns, table, schema
        )

        for index in range(len(data)):
            row = data[index]

            insert_sql_value = self._generate_insert_sql_value(row)
            insert_sql = insert_sql_prefix + insert_sql_value
            insert_sql_size = len(to_utf8_bytes(insert_sql))
            if insert_sql_size >= self._MAX_STATEMENT_SIZE:
                raise DataEncodeError(
                    "Row[{}] is beyond size limitation".format(index)
                )

    def _generate_insert_sql_prefix(self, columns, table, schema):
        encode = self._encode_metadata

        insert_sql_prefix = "INSERT INTO {}.{} ({}) VALUES ".format(
            encode(schema),
            encode(table),
            ",".join([encode(column) for column in columns]),
        )

        return insert_sql_prefix

    def _generate_insert_sql_value(self, row):
        return "({})".format(",".join(row))

    def _generate_insert_sqls(self, data, columns, dtype, table, schema):
        insert_sql_prefix = self._generate_insert_sql_prefix(
            columns, table, schema
        )

        remaining_rows = len(data)
        total_rows = remaining_rows

        insert_sql = insert_sql_prefix
        for index in range(total_rows):
            row = data[index]
            insert_sql_value = self._generate_insert_sql_value(row)
            if index == 0:
                insert_sql = insert_sql_prefix + insert_sql_value
                continue
            tmp_insert_sql = insert_sql + "," + insert_sql_value
            statement_size = len(to_utf8_bytes(tmp_insert_sql))
            if statement_size >= self._MAX_STATEMENT_SIZE:
                remaining_rows = max(0, total_rows - index)
                yield remaining_rows, insert_sql + ";"
                insert_sql = insert_sql_prefix + insert_sql_value
            else:
                insert_sql = tmp_insert_sql
        yield 0, insert_sql + ";"

    def _load_chunks(self, data, columns, dtype, table, schema):
        if len(data) == 0:
            return

        insert_sqls = self._generate_insert_sqls(
            data, columns, dtype, table, schema
        )

        statement_ids, run_sqls = [], []
        for remaining_rows, insert_sql in insert_sqls:
            yield remaining_rows

            run_result = self.run_query(insert_sql)

            statement_id = run_result["Id"]
            statement_ids.append(statement_id)
            run_sqls.append(insert_sql)

        self.wait(run_sqls, statement_ids)

    def load_dataframe(self, data, columns, dtype, table, schema="public"):
        total_rows = len(data)
        progress = self._load_chunks(data, columns, dtype, table, schema)
        for remaining_rows in progress:
            logger.info(
                "{} out of {} rows loaded.".format(
                    total_rows - remaining_rows, total_rows
                )
            )

    def exists_table(self, table, schema="public", database=None):
        table_map = self._get_table_map(schema, database)
        return table in table_map

    def _get_table_map(self, schema, database):
        # in this method, it is possible to list tables for another database
        # when listing tables for another table, should specify the parameter
        # database
        list_tables_params = {"ClusterIdentifier": self.cluster_identifier}
        # if `database` param is skipped, database is designated as the same
        # as connected database
        list_tables_params["Database"] = (
            database if database is not None else self.database
        )
        list_tables_params["SchemaPattern"] = schema
        # add authentication info
        if self.secret_arn is not None:
            list_tables_params["SecretArn"] = self.secret_arn
        else:
            list_tables_params["ConnectedDatabase"] = self.database
            list_tables_params["DbUser"] = self.db_user

        table_map = set()
        while self._has_next_token(list_tables_params):
            list_tables_result = self.client.list_tables(**list_tables_params)
            for table_info in list_tables_result["Tables"]:
                table_map.add(table_info["name"])
            list_tables_params["NextToken"] = list_tables_result.get(
                "NextToken", self._NO_NEXT_TOKEN
            )
        return table_map



from .core import read_redshift, to_redshift

from .types import (
    RedshiftType,
    DoublePrecision,
    Real,
    Numeric,
    Integer,
    SmallInt,
    BigInt,
    Boolean,
    Char,
    BPChar,
    VarChar,
    Text,
    TimeStamp,
    TimeStampTz,
    Date,
    Time,
    TimeTz,
    Geometry,
    Super,
)

__all__ = [
    "read_redshift",
    "to_redshift",
    "RedshiftType",
    "DoublePrecision",
    "Real",
    "Numeric",
    "Integer",
    "SmallInt",
    "BigInt",
    "Boolean",
    "Char",
    "BPChar",
    "VarChar",
    "Text",
    "TimeStamp",
    "TimeStampTz",
    "Date",
    "Time",
    "TimeTz",
    "Geometry",
    "Super",
]

from . import _version

__version__ = _version.get_versions()["version"]

del _version



# Copyright 2019. PyData Development Team
# Distributed under BSD 3-Clause License.
# See LICENSE.txt for details.

"""Simple upload example."""

import argparse

import sys

sys.path.append("..")


def main(
    cluster_identifier,
    database,
    table,
    secret_arn,
):
    # [START pandas_amazon_redshift_to_redshift_simple]
    import pandas
    import pandas_amazon_redshift

    # TODO: Set cluster_identifier to specify your Redshift cluster.
    # cluster_identifier = "my-cluster"

    # TODO: Set the name of database to store your data.
    # database = "my_database"

    # TODO: Set the name of table to store your data.
    # table = "my_table"

    # TODO: Set database user name or secret ARN to access the database above.
    #       In this sample, we use ``secret_arn``.
    # secret_arn = """
    # arn:aws:secretsmanager:<aws-region>:<aws-account-id>:secret:<secret-name>
    # """

    df = pandas.DataFrame(
        {
            "my_string": ["a", "b", "c"],
            "my_int64": [1, 2, 3],
            "my_float64": [4.0, 5.0, 6.0],
            "my_bool1": [True, False, True],
            "my_bool2": [False, True, False],
            "my_dates": pandas.date_range("now", periods=3),
        }
    )

    dtype = {
        "my_string": "CHAR",
        "my_int64": "BIGINT",
        "my_float64": "DOUBLE PRECISION",
        "my_bool1": "BOOLEAN",
        "my_bool2": "BOOLEAN",
        "my_dates": "DATE",
    }

    pandas_amazon_redshift.to_redshift(
        df, cluster_identifier, database, table, dtype, secret_arn=secret_arn
    )
    # [END pandas_amazon_redshift_to_redshift_simple]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cluster-identifier")
    parser.add_argument("--database")
    parser.add_argument("--secret-arn")
    parser.add_argument("--table")
    args = parser.parse_args()
    main(**vars(args))



# Copyright 2021. Yutaka Kabutoya
# Distributed under BSD 3-Clause License.
# See LICENSE.txt for details.

"""Simple query example."""

import argparse

import sys

sys.path.append("..")


def main(
    cluster_identifier,
    database,
    db_user,
):
    # [START pandas_amazon_redshift_read_redshift_simple]
    import pandas_amazon_redshift

    # TODO: Set cluster_identifier to specify your Redshift cluster.
    # cluster_identifier = "my-cluster"

    # TODO: Set the name of database where your data is stored.
    # database = "my_database"

    # TODO: Set database user name or secret ARN to access the database above.
    #       In this sample, we use ``db_user``.
    # db_user = "my_db_user"

    sql = """
    SELECT sum(qtysold)
    FROM   sales, date
    WHERE  sales.dateid = date.dateid
    AND    caldate = '2008-01-05';
    """
    df = pandas_amazon_redshift.read_redshift(
        sql, cluster_identifier, database, db_user=db_user
    )
    # [END pandas_amazon_redshift_read_redshift_simple]
    print(df)
    return df


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cluster-identifier")
    parser.add_argument("--database")
    parser.add_argument("--db-user")
    args = parser.parse_args()
    main(**vars(args))



class InvalidAuthentication(ValueError):
    """Raised when no info for temporary credentials or secret arn."""

    pass


class QueryFailedError(ValueError):
    """Raised when query is failed."""

    pass


class QueryAbortedError(ValueError):
    """Raised when query is aborted."""

    pass


class TableCreationError(ValueError):
    """Raised when the create table method fails"""

    pass


class DataEncodeError(ValueError):
    """Raised when the data encode fails"""

    pass


class MetadataEncodeError(ValueError):
    """Raised when the metadata escape fails"""

    pass


class InvalidRedshiftType(ValueError):
    """Raised when Redshift type is designated with unexpected string."""

    pass


class NoColumnError(ValueError):
    """Raised when input dataframe has no columns"""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import versioneer
from setuptools import find_packages, setup

NAME = "pandas-amazon-redshift"


# versioning
cmdclass = versioneer.get_cmdclass()


def readme():
    with open("README.rst") as f:
        return f.read()


INSTALL_REQUIRES = [
    "setuptools",
    "pandas>=1.3.0",
    "boto3>=1.17.100",
    "botocore>=1.20.100",
]

setup(
    name=NAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Pandas interface to Amazon Redshift",
    long_description=readme(),
    license="BSD License",
    author="Yutaka Kabutoya",
    author_email="y.kabutoya@gmail.com",
    url="https://github.com/yutaka-kabu/pandas-amazon-redshift",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
    keywords="data",
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.7.1",
    packages=find_packages(exclude=["boto3_mock*", "tests*"]),
    test_suite="tests",
)
